/******************************************
HKUST ELEC254 Project - Summer 2008
Group #4: Wang Zhen, Zhu Shucheng, Zhuo Wei
Author: Wang Zhen
******************************************/

#include "keypad.h"

unsigned char Key;							// store the pressed key
unsigned char KeyBuffer[KeyBufferSize];		// store the scan code
unsigned char KeyBufferStart;					// start point of the buffer
unsigned char KeyBufferEnd;					// end point of the buffer
bit Flag_KeypadLED;							// light -> 1; dark -> 0

//function definition
void keypad_initial()
{
    IT0 = 1;	//falling-edge triggered external interrupt 0 (1-0 transition)
    EX0 = 1;	//enable external interrupt 0
    EA = 1;     //enable the interrupt
    
    // idle signal for receving data
    Clock_Pin = 1;
    Data_Pin = 1;
    
    //setting
    Key = 0;
    Flag_KeypadLED = 0;
    KeyBufferStart=0;
    KeyBufferEnd=0;
    TurnOnLED();
    
    return;
}

void keypad_interrupt() interrupt 0
{
	static unsigned char ScanCode = 0;	// store the scan code
	static unsigned char bitcount = 0;	// keep track of the bits received
	static bit parity = 0;				// trace the parity: odd -> 1; even -> 0;
	static bit data_ready = 0;			// data is ready -> 1; data is not ready -> 0

    ++bitcount;
    //start bit
    if(bitcount == 1)
    {
        if(Data_Pin)    //error: start bit should be 0
        {
            bitcount = 0;
            return;
        }
    }
    //data bits
    else if(bitcount >= 2 && bitcount <= 9)
    {
        ScanCode = ScanCode * 2;  //left shift by 1 bit
        if(Data_Pin)
        {
            ScanCode = ScanCode + 1;
            if(parity)
                parity = 0;
            else
                parity = 1;
        }
    }
    //parity bit
    else if(bitcount == 10)
    {
        if(parity == Data_Pin)  //data error
            data_ready = 0;     //data is not ready
        else                    //correct data
            data_ready = 1;     //data is ready
    }
    //stop bit
    else if(bitcount == 11)
    {
        if(!data_ready)     //stop bit error
            ScanCode = 0;
        
		//store the scan code into the buffer
		KeyBuffer[KeyBufferEnd] = ScanCode;
		KeyBufferEnd = (KeyBufferEnd + 1) % KeyBufferSize;
		if(KeyBufferEnd==KeyBufferStart)
		{
			if(KeyBufferEnd==0)
				KeyBufferEnd = KeyBufferSize - 1;
			else
				KeyBufferEnd--;
		}

		//reset
        bitcount = 0;
        parity = 0;
        data_ready = 0;
    }
    //error counting
    else
    {	//reset
		ScanCode = 0;
        bitcount = 0;
        parity = 0;
        data_ready = 0;
    }
    return;
}

unsigned char Keypad_Decoder()
{
	static unsigned char is_up=0;
	unsigned char sc;

	//empty KeyBuffer
	if(KeyBufferStart==KeyBufferEnd)
		return Key;

	//get the scan code
	sc = KeyBuffer[KeyBufferStart];
	KeyBufferStart = (KeyBufferStart + 1) % KeyBufferSize;

	//start to decode
	if (!is_up)
	{
		if(sc == 0x0F)			// The up-key identifier
		{
			is_up = 1;
			return Key;
		}
		else if(sc == 0x07)		// extended key identifier ignored
		{
			return Key;
		}
		else if(sc==0xEE)		//NUM-LOCK
		{
			if(Key!=0xEE)
				Keypad_LED();
			return sc;
		}
		else
		{
			if(Flag_KeypadLED)
				return sc;
			else
			{
				// /, *, -, +, Enter
				if((sc==0x5A)||(sc==0x52)||(sc==0x3E)||(sc==0xDE)||(sc==0x9E))
					return sc;
				//others
				else
					return sc + 1;
			}
		}
	}
	else //break-code: reset Key
	{
		is_up = 0;
        return 0;
	}
}

void KeypadCommand(unsigned char Command)
{
    unsigned char CommandCount;
    bit LEDparity = 0;  //even -> 0; odd -> 1
    EX0 = 0;            //disable external interrupt 0
    
    Clock_Pin = 0;      //bring the clock line low for at least 100us
    delay_ms(1);
    
    //start bit
    Data_Pin = 0;       //bring the data line low
    Clock_Pin = 1;      //release the clock line
    while(Clock_Pin){}  //wait for the falling edge
    
    //8 data bits
    for(CommandCount = 8; CommandCount > 0; CommandCount--)
    {
        if((Command & 0x01) == 0x01)    //check the last bit of the command
        {
            Data_Pin = 1;
            if(LEDparity)               //trace the parity
                LEDparity = 0;
            else
                LEDparity = 1;
        }
        else
            Data_Pin = 0;
        
        Command = Command / 2;          //shift the command to the left
        while(!Clock_Pin){}             //wait for the high level status
        while(Clock_Pin){}              //wait for the falling edge
    }
    //parity bit
    if(LEDparity)
        Data_Pin = 0;
    else
        Data_Pin = 1;
    while(!Clock_Pin){}                 //wait for the high level status
    while(Clock_Pin){}                  //wait for the falling edge
    //stop bit
    Data_Pin = 1;                       //release the data line
    while(!Clock_Pin){}                 //wait for the high level status
    while(Clock_Pin){}                  //wait for the falling edge
    //Ack
    while(!Clock_Pin){}                 //wait for the high level status
    while(Clock_Pin | Data_Pin){}       //wait for clock -> 0; data -> 0
    //over
    while(!(Clock_Pin & Data_Pin)){}    //wait for clock -> 1; data -> 1
    
    delay_ms(1);
    EX0 = 1;                            //re-enable external interrupt 0
}

void TurnOnLED()
{
    KeypadCommand(0xED);
    KeypadCommand(KeypadLEDon);
    Flag_KeypadLED = 1;
}

void TurnOffLED()
{
    KeypadCommand(0xED);
    KeypadCommand(KeypadLEDoff);
    Flag_KeypadLED = 0;
}

void Keypad_LED()
{
	if(Flag_KeypadLED)
		TurnOffLED();
	else
		TurnOnLED();
}
